#!/usr/bin/perl

##########################################################################################
#	  This file is part of Proteinortho.
#	  (C) 2009/2010 Marcus Lechner
# 
#	  Proteinortho is free software; you can redistribute it and/or modify
#	  it under the terms of the GNU General Public License as published
#	  by the Free Software Foundation; either version 2, or (at your
#	  option) any later version.
#
#	  Proteinortho is distributed in the hope that it will be useful, but
#	  WITHOUT ANY WARRANTY; without even the implied warranty of
#	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#	  General Public License for more details.
#
#	  You should have received a copy of the GNU General Public License
#	  along with Proteinortho; see the file COPYING.  If not, write to the
#	  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#	  Boston, MA 02111-1307, USA.	
##########################################################################################

##########################################################################################
# About
##########################################################################################
# Proteinortho
# input fasta files with proteins
# output matrix with orthologous proteins
# 
# @author Marcus Lechner
# @email marcus@bioinf.uni-leipzig.de
# @company Bioinformatics, University of Leipzig
# @date 18-10-2010
#
##########################################################################################

##########################################################################################
# Imports
##########################################################################################
use strict;
use warnings "all";
use File::Basename;
use threads;
use threads::shared;
use vars qw/@query $cpu $debug $blastoutpath $force $verbose $fforce $keep $blastonly %seqs/;

##########################################################################################
# Install signal handler
##########################################################################################
$SIG{'INT'}  = 'sighandler';
$SIG{'TERM'} = 'sighandler';
##########################################################################################

##########################################################################################
# Usage, defaults and parameters
##########################################################################################
my $usage= << "JUS";
  usage: proteinortho4.pl [OPTIONS] FASTA1 FASTA2 [FASTA...]
  or:    proteinortho4.pl [OPTIONS] LIST_OF_FASTAS
  needs: blastall, formatdb
version: 4.18
options: -e=          E-value for blast [default: 1e-10]
         -p=          blast program {blastn|blastp} [default: blastp]
         -id=         min. percent identity of best blast hits [default: 25]
         -cov=        min. coverage of best blast hits [default: 0.5]
         -conn=       min. algebraic connectivity [default: 0.1]
         -m=          min. similarity for additional hits (0..1) [default: 0.95]
         -pairs       do not remove simple pairs from output   
         -selfblast   apply selfblast, detects paralogs without orthologs
         -unambiguous exclude connected components with paralogs from output
         -a=          number of processors to use [default: auto]
         -f           force blastall (even if blast output is found)
         -ff          force formatdb (even if databases are found), implies -f
         -dir=        blast output directory (default is the working directory)
         -remove      removes outputs from blast output directory after use
         -batch       returns a batch-list of jobs on STDOUT
                      this is preferable to -blastonly if you use a
                      cluster-management system. "wait;" will indicate that all
                      jobs above have to be finished before proceeding further
                      requires the -o= parameter for final output
         -blastonly   only do the blasts: capable to run on multiple machines
                      synchronized blast output directory,
                      allows clean break with 'touch stop' or SIGTERM
         -blastdone   use after a complete run with -blastonly
		      does not check output files, just runs the algorithm
		      on previously created blast output
	 -log=	      writes detailed log of reciprocal best hits to file
	 -o=	      prints output to the given file rather than STDOUT
         -verbose     keeps you informed about the progress
         -debug       gives detailed information for bug tracking
                      does not work with -verbose

                      one entry per in line in LIST_OF_MULTIFASTAS
                      check the manpages for more detailed information

------------------------------------------------------------------------------
!IMPORTANT! ensure that your protein IDs are unique ACROSS species !IMPORTANT!
------------------------------------------------------------------------------
JUS

my @options = grep {substr($_, 0, 1) eq '-'} @ARGV;
my @others  = grep {substr($_, 0, 1) =~ /^[^-]/} @ARGV;

## default settings:
my $evalue = "1e-10";
my $blast = "blastp";
my $unambigous = 0;
my $noselfb = 1;
my $maxl = 0.95;
my $bhold = 1;
my $blastdone = 0;
my $ident = 25;
my $log = "";
my $outfile = "";
our $batch = 0;
my $cov=0.5;
my $inc_pairs = 0;
our $iolimit = 1;
our @joinable = (); # 4threading
share(@joinable);   # 4threading
$cpu = 0;           # 4threading
share($cpu);
$keep = 1;
$force=0;
$debug=0;
$fforce=0;
our $con = -1;
$verbose=0;
$blastonly = 0;
our $gotlock = 0;	# Makes shure lock wont be stolen from another process
my $files;

## setting parameters:
foreach my $option (@options){
  if    ($option =~ m/^-remove$/)     { $keep = 0;   }
  elsif ($option =~ m/^-blastonly$/)  { $blastonly = 1;   }
  elsif ($option =~ m/^-blastdone$/)  { $blastdone = 1; }
  elsif ($option =~ m/^-verbose$/)    { $verbose = 1;  }
  elsif ($option =~ m/^-debug$/)      { $debug = 1;  }
  elsif ($option =~ m/^-selfblast$/)  { $noselfb = 0;  }  
  elsif ($option =~ m/^-batch$/)      { $batch = 1; }
  elsif ($option =~ m/^-f$/)          { $force = 1;  }
  elsif ($option =~ m/^-ff$/)         { $fforce = 1; $force = 1; }
  elsif ($option =~ m/^-p=(.*)$/)     { $blast = $1; }
  elsif ($option =~ m/^-e=(.*)$/)     { $evalue= $1; }
  elsif ($option =~ m/^-a=(\d*)$/)    { $cpu   = $1; }
  elsif ($option =~ m/^-a=auto$/)     { $cpu   = 0; }
  elsif ($option =~ m/^-noiolimit$/)     { $iolimit   = 0; }
  elsif ($option =~ m/^-m=(\d+.?\d*)$/)    { $maxl  = $1; }
  elsif ($option =~ m/^-unambiguous$/)    { $unambigous  = 1; }
  elsif ($option =~ m/^-dir=(.+)$/)   { $blastoutpath  = $1; }
  elsif ($option =~ m/^-o=(.+)$/)     { $outfile  = " >$1"; }
  elsif ($option =~ m/^-id=([0-9]+)$/)     { $ident = $1; }
  elsif ($option =~ m/^-conn=([0-9\.]+)$/)     { $con = $1; }
  elsif ($option =~ m/^-cov=([0-9\.]+)$/)     { $cov = $1; }
  elsif ($option =~ m/^-log=(.+)$/)     { $log = $1; }
  elsif ($option =~ m/^-pairs$/)     { $inc_pairs = 1; }
  else  { die "Invalid command line option: \'$option\'!\n$usage"; }
}

if ($debug && $verbose) {
	die "Sorry, -verbose and -debug cannot be used together!\n$usage"; 
}

if ($cpu == 0) {
	print STDERR "Detecting CPU cores...";
	my $cpu_x = qx(grep processor /proc/cpuinfo | wc -l);
	$cpu_x =~ s/[^0-9]//g;
	if (length($cpu_x) == 0 || $cpu_x == 0) {print STDERR "failed! Use 1 core only\n";$cpu_x = 1;}
	print STDERR "$cpu_x cores detected\n";
	$cpu = $cpu_x;
}

print STDERR "Detecting blastall...";
my $blastversion = "";
{my @blastv = qx("blastall");
foreach (@blastv) {
	chomp;
	if ($_ =~ /blastall.+?([^\s]+)/) {
		$blastversion = $1;
		last;
	}
}
}
if ($blastversion eq "") {
	print STDERR "failed! 'blastall' does not seem to be working\n";
}
print STDERR "ok (version $blastversion)\n";

# a list can also be given
# if only one argument is given
if (scalar(@others)==1) {
	$files = "-list $others[0]";
	# open the (possible) list 
	open(LIST,"<$others[0]") || die ("Could not find MULTIFASTA-list. \n$usage");
	@others = <LIST>;
	chomp(@others);
	close(LIST);
	# use it's lines as arguments
	foreach my $line (@others) {
		unless ($line =~ /.\../) {
			die ("$line in $others[0]\n does not seem to be a file!\nAt least two sequence files expected: MULTIFASTA1 MULTIFASTA2!\n\n$usage");
		}
	}
}
else {
	$files = join(" ",@others);
}

# still we need at least two
die "At least two sequence files expected: MULTIFASTA1 MULTIFASTA2!\n\n$usage"
    unless scalar(@others)>=2;

# bad parameter
if ($maxl < 0 || $maxl > 1) {
	die "Value for parameter -m must be within 1 and 2!\n\n$usage";
}

# bad parameter combination
if (($blastonly == 1 || $batch == 1) && $keep == 0) {
	die "Combination of remove and -blastonly/-batch does not make much sense!\n\n$usage";
}

if ($blastonly == 1 && $batch == 1) {
	die "Blastonly and -batch cannot be used together!\n\n$usage";
}

if ($batch == 1 && $outfile eq "") {
	die "-batch requires -o= parameter!\n\n$usage";
}

# check file accessibility
print STDERR "Checking accessibility of files... \n";
for (my $i=0; $i<scalar(@others); $i++) {
  die "Can\'t read file \'$others[$i]\'!\n"
      unless open(FH, $others[$i]);
  close FH;
}

##########################################################################################
# Main programm
#
##########################################################################################

##########################################################################################
# Step 1: format input-databases
##########################################################################################
my $stat_runs = scalar(@others)*(scalar(@others)-$noselfb)/2;
if (!$noselfb) {$stat_runs = ($stat_runs*2+scalar(@others))/2;}

my $glob_counter = 0;
my $stime;

# don't do if blast was already done
unless ($blastdone) {

### Batched ###
if ($batch) {
	print STDERR "Generating batch job...\n";
	for (my $i=0; $i<scalar(@others); $i++) {
		&formatdb($others[$i],$blast);
	}
	print "wait;\n";
}
###############
else {


print STDERR "Formating databases...\n";
# for each given fasta-file
for (my $i=0; $i<scalar(@others); $i++) {
  if ($force == 1 || do_formatdb($others[$i],$blast) == 1) {
	# Wait for free Thread-Slots
	my @tliste = threads->list();
	while(defined($tliste[0]) && scalar@tliste >= $cpu*1.5) {
		sleep(1);
		if (scalar(@joinable) > 0) {
			lock(@joinable);

			# Fetch finished threads
			while(scalar@joinable>0) {
				my $tid = shift(@joinable);
				my $thr = threads->object($tid);
				$thr->join();
				@tliste = threads->list();
			}
		}
	}

	# formatdb
	my $dbthread = threads->create('formatdb', ($others[$i],$blast));
  }
  else {
	if ($debug) {print STDERR " Formated database for '$others[$i] exists'\n";}
  }
}

# Gather unfinished threads
foreach my $dbthread (threads->list()) {
	$dbthread->join();
}
@joinable = ();
} # Batched
##########################################################################################
# Step 2: run the blasts (the biggest part)
##########################################################################################
# define blastoutpath
unless (defined($blastoutpath)) {
#	my $prepath=qx('pwd');
#	chomp($prepath);
	$blastoutpath=".";
}
# Remove slash in the end
#$blastoutpath =~ s/\/$//; 

### Batched ###
if ($batch) {
	if ($debug) {print STDERR "Generating batch job...\n";}
	for (my $i=0; $i<scalar(@others)-$noselfb; $i++) {
		for (my $j=$i+$noselfb; $j<scalar(@others); $j++) {
			&blast($others[$i], $others[$j], $blast, $evalue, $bhold, $maxl);
		}
	}
	print "wait;\n";
}
else {
###############
print STDERR "Blasting (this may take some time)...\n";

# Syncronisation for blastonly
my $s_i = 0; # only needed for sync in blastonly case
my $s_j = 0;
if ($blastonly == 1) {
	# init syncfile
	unless (-e $blastoutpath."/sync") {
		print STDERR "Syncfile not found. Adding new one: ".$blastoutpath."/sync\n";
		&fileLock;
		open(SYNC,">$blastoutpath"."/sync") || die "sync: $!";
		print SYNC "0 0";
		close(SYNC);
		&fileUnlock;
		print STDERR "Sync creation done\n";
	}
	else {
		print STDERR "Syncfile found. Sycronizing...\n";
	}
	# read current progress from syncfile
	($s_i, $s_j) = &sync();
	$glob_counter = ($s_i-1*$s_i)/2+$s_j-$s_i;
}

# start the blastruns
if ($verbose) {$|++;}
$stime = time;
my $runs_done = 0;
for (my $i=$s_i; $i<scalar(@others)-$noselfb; $i++) {			# -1
  for (my $j=$i+$noselfb; $j<scalar(@others); $j++) {			#+1
	$glob_counter++;

	# Syncronisation for blastonly
	if ($blastonly == 1) { 
		# skip what others should have done
	  	if ($i < $s_i || ($i == $s_i && $j <= $s_j)) {next;}
		
		# tell all I do a the job
	    	($s_i, $s_j) = &sync($i,$j); 		
		# stop order received, don't start new runs
	    	if ($s_i == -1) {$i = scalar(@others);last;} 
		
		
	    	unless ($i > $s_i || ($i == $s_i && $j > $s_j)) {next;}	 # skip what others have
	}

    ## Precheck if files already exsist, saves waiting for threadslots in this case
    if (&blast_needed($others[$i], $others[$j])) {

    # Wait for free Thread-Slots
    my @tliste = threads->list();
	while(defined($tliste[0]) && scalar@tliste >= $cpu) {
		sleep(1);
		if (scalar(@joinable) > 0) {
			lock(@joinable);
			# Fetch finished threads
			while(scalar@joinable>0) {
				my $tid = shift(@joinable);
				my $thr = threads->object($tid);
				$thr->join();
				if ($verbose) {
					if ($blastonly) {
						print STDERR (int($glob_counter*1000/$stat_runs)/10)."% (".(int(++$runs_done*1000/$stat_runs)/10)."% by this process)   \r";				
					}
					else {
						print STDERR (int(++$runs_done*1000/$stat_runs)/10)."%   \r";
					}				
				}
				@tliste = threads->list();
			}
		}
	}
	my ($blastthread) = threads->create('blast', ($others[$i], $others[$j], $blast, $evalue, $bhold, $maxl));
    } # if (&blast_needed)
    elsif ($debug) {	
	    print STDERR " $others[$i] vs. $others[$j] \n-> blast output exists.\n\n";
	    print STDERR " $others[$j] vs. $others[$i] \n-> blast output exists.\n\n";
    }
  }

}

# Gather unfinished threads
foreach my $thr (threads->list()) {
	$thr->join();
	if ($verbose) {
		if ($blastonly) {
			print STDERR (int($glob_counter*1000/$stat_runs)/10)."% (".(int(++$runs_done*1000/$stat_runs)/10)."% by this process)   \r";				
		}
		else {
			print STDERR (int(++$runs_done*1000/$stat_runs)/10)."%   \r";
		}				
	}
}
@joinable = ();

if ($verbose) {
		print STDERR "                  \r";
		if ($blastonly) {
			print STDERR (int($glob_counter*1000/$stat_runs)/10)."% (".(int(++$runs_done*1000/$stat_runs)/10)."% by this process)   \n";				
		}
		else {
			print STDERR (int($glob_counter*1000/$stat_runs)/10)."%   \n";
		}				
		$|--;
}

if ($s_i == -1) { print STDERR "\nCalculation terminated due to existance of stop-file\n";}

} # batched

if ($blastonly == 1 && !$batch) {
  print STDERR "Blastoutput generated\nFinish by using -blastdone after all processes on all nodes have finished.";
  exit;
}
} # blastdone
elsif(!$batch) {
	print STDERR "Running in mode -bastdone. Blast output should already exist!\n";
}


##########################################################################################
# Step 3: Analyze the blast-output 
##########################################################################################
my @tmppath = fileparse($0); # path to the C++-part of this program
my $options = "-evalue $evalue -m $maxl ";
if ($blastoutpath ne ".") {
	$options .= "-blastdir $blastoutpath ";
}
if ($noselfb == 0) {
	$options .= "-selfblast ";
}
if ($unambigous == 1) {
	$options .= "-unambigous ";
}
if ($con >= 0) {
	$options .= "-connectivity $con ";
}
if ($cov >= 0) {
	$options .= "-coverage $cov ";
}
if ($iolimit == 0) {
	$options .= "-noiolimit ";
}
if ($verbose) {
	$options .= "-verbose ";
}
if ($debug) {
	$options .= "-debug ";
}
if ($inc_pairs) {
	$options .= "-include_pairs ";
}
if ($log ne "") {
	$options .= "-log $log ";
}

$options .= "-blastversion \"$blast v$blastversion\" ";

$options .= "-threads $cpu ";

$options .= "-percent_identity $ident ";

# print "$tmppath[1]/po_worker $options".join(" ",@others),"\n";

if ($batch) {
	print "$tmppath[1]/portho4_worker $options".join(" ",@others).$outfile."\n";
	print STDERR "Batch job written to STDOUT\n";	
	exit;
}

my $return = system("$tmppath[1]/portho4_worker ".$options.$files.$outfile);

##########################################################################################
# Step 4: Cleanup
##########################################################################################
if ($keep == 0) {
	print STDERR "Removing unnessesary files...\n";
	for (my $i=0; $i<scalar(@others); $i++) {
		if ($blast eq "blastn") {
			system("rm $others[$i].nhr $others[$i].nin $others[$i].nsq");
		}
		elsif ($blast eq "blastp") {
			system("rm $others[$i].phr $others[$i].pin $others[$i].psq");
		}
  	}

	for (my $i=0; $i<scalar(@others)-$noselfb; $i++) {			# -1
		for (my $j=$i+$noselfb; $j<scalar(@others); $j++) {			#+1
		  	my ($path1,$name1)=file2path_name($others[$i]);
			my ($path2,$name2)=file2path_name($others[$j]);
			unlink($blastoutpath."/".$name2.".".$name1.".bla");	
			unlink($blastoutpath."/".$name1.".".$name2.".bla");
		}
	}
}

##########################################################################################
# FINISH, we are done
##########################################################################################
exit $return;




























##########################################################################################
# Subroutines
#
##########################################################################################

##########################################################################################
# Fasta-Format
##########################################################################################

# checks if formatdb is needed (unneeded, if the files already exist)
# make shure the path-detecion is the same as in sub formatdb
sub do_formatdb {
  my($faa,$blast)=@_;
#	print STDERR "File: $faa\n";
  my $val=1;
  if ($blast eq "blastn") {
    if (-e "$faa.nhr" &&
	-e "$faa.nin" &&
	-e "$faa.nsq") {
      $val=0;
    }
  }
  elsif ($blast eq "blastp") {
    if (-e "$faa.phr" &&
	-e "$faa.pin" &&
	-e "$faa.psq") {
      $val=0;
    }
  }
  else {
    die "$blast is not supported yet! Feel free to adapt the code to your needs. The programmer.\n"
  }
  return $val;
}

# Run format-db (THREAD)
sub formatdb {
  my($faa,$blast)=@_;
##(1) format multi-fasta files with formatdb
#------------------------------------------------------------------------------
  if ($blast =~ m/^blastp$/i) {
    if ($batch) {
	print "formatdb -i $faa -p T -o F\n";
    }
    else {
        if ($debug) {print STDERR " Formating database '$faa'...\n";}
	`formatdb -i $faa -p T -o F`;
    }
  }
  elsif ($blast =~ m/^blastn$/i) {
    if ($batch) {
	print "formatdb -i $faa -p F -o F\n";
    }
    else {    
        if ($debug) {print STDERR "Formating database '$faa'...\n";}
	`formatdb -i $faa -p F -o F`;
    }
  }
  else {
    die "$blast is not supported yet! Feel free to adapt the code to your needs. The programmer.\n"
  }
#------------------------------------------------------------------------------

  # note, this thread has finished
  lock(@joinable);
  push(@joinable,threads->tid());
  return;
}

##########################################################################################
# Blast
##########################################################################################
# checks if a blast is needed (unneeded, if the files already exist)
# make shure the path-detecion is the same as in sub blast
sub blast_needed {
  my ($faa1, $faa2)=@_;
  my ($path1,$name1)=file2path_name($faa1);
  my ($path2,$name2)=file2path_name($faa2);

  my $blastout1=$blastoutpath."/".$name1.".".$name2.".bla";
  my $blastout2=$blastoutpath."/".$name2.".".$name1.".bla";

  if ($force == 0 && -e "$blastout1" && -e "$blastout2") {
	return 0;
  }
  return 1;
}

# prepares and runs the blasts (THREAD)
sub blast {
  my ($faa1, $faa2, $blast, $evalue, $bhold, $maxl)=@_;

  # gather paths
  my ($path1,$name1)=file2path_name($faa1);
  my ($path2,$name2)=file2path_name($faa2);
 
  #do blastall: faa1 as query, faa2 as database
  my $blastout1=$blastoutpath."/".$name1.".".$name2.".bla";
  if ($debug && !$batch) {print STDERR "Generating blast output: '$faa1' vs. '$faa2'...\n";}
  blastall($faa1,$faa2,$blast,$evalue,$blastout1);

  unless ($faa1 eq $faa2) {		# 3.1.1 bidirectional not needed
	  #do blastall: faa2 as query, faa1 as database
	  my $blastout2=$blastoutpath."/".$name2.".".$name1.".bla";
	  if ($debug && !$batch) {print STDERR "Generating blast output: '$faa2' vs. '$faa1'...\n";}
	  blastall($faa2,$faa1,$blast,$evalue,$blastout2);
  }

  # note, this thread has finished
  lock(@joinable);
  push(@joinable,threads->tid());

  return;
}


# does the blast and writes it to harddrive
sub blastall {
  my ($query, $db, $blast, $evalue, $out)=@_;
  if ($batch) {
	print "blastall -d $db -i $query -p $blast -a 1 -m8 -e $evalue >$out\n";
  }
  else {
	my $stderr = " 2>/dev/null";
	if ($debug) {$stderr = "";}
  	`blastall -d $db -i $query -p $blast -a 1 -m8 -e $evalue >$out$stderr`;
  }
}

# divides a string into filename and it's path
sub file2path_name {
  my $file =shift;
  my ($name, $path) = fileparse($file);
  return $path, $name;
}

##########################################################################################
# Syncronising (used only if blastonly parameter is set)
##########################################################################################

# tells if blast-job could done next, or was already taken by another process
# basic syncronising method, does locking on it's own
sub sync {
    # the requested job
    my $i = shift;	
    my $j = shift;

    # if the (file based) stop-signal is found, we will stop
	if (-e $blastoutpath."/stop") {
		print STDERR "Recived order to terminate procedure...\n Finishing running threads\n";
		return(-1,-1);
	}

      # lock$evalue
	&fileLock;

	# open the syncfile
	open(SYNC,"+<","$blastoutpath"."/sync") || die "sync: $!";
	my @content = <SYNC>;
	chomp(@content);
	# gather current job
	(my $s_i, my $s_j) = split(/ /,$content[0],2);
	if (defined($i) && defined($j)) {
		# if we do the job, update sync-data	
		if ($i > $s_i || ($i == $s_i && $j > $s_j)) {
			seek(SYNC,0,0);
			truncate(SYNC, 0);
			print SYNC "$i $j";
			
			# inform the user
			if ($debug) {print STDERR "Taking Job No $i x $j\n";}
		}
	}
	close(SYNC);

	# unlock	
	&fileUnlock;

	# return the next job to be done
	return ($s_i, $s_j);
}

# locks the sync-file
# always has to be done, before reading or writing it
# don't forget to unlock afterwards
sub fileLock {
	my $counter = 0;
	# if there already is a lock, wait
      # for flock does not work within NFS, we simulate it
      # with a directory to stay comaptible
	while (!mkdir($blastoutpath."/lock",0755)) {   
		# sleep for 1 sec and try again
		sleep(1);     

		# if access failed to often, there might be a problem
		$counter++;
		if ($counter > 30) {
			die("Termination due to locking problem. Directory $_[0]");
		}$maxl
     }

     # note globaly, that we've got the lock
     $gotlock = 1;

     # if the locking took longer, note it within the error-stream
     if ($debug && $counter > 0) {
	   	print STDERR "Got lock on sync after $counter trys\n";
     }
}

# unlocks the sync-file
sub fileUnlock {
	# only remove the lock, if it's mine
	if ($gotlock == 0) {return;}	 

	# remove file lock dir
	rmdir($blastoutpath."/lock");     

	# note globaly, that we don't have the lock anymore
      $gotlock = 0;
}

# is called when the signals int term are caught
# shuts down the process safely, even if more of them are running syncronized
# finishes all threads, removes it's lock and exits
sub sighandler {
    # read signal number
    my $signal = shift;

    # reinstall signal-handler
    $SIG{'INT'}  = 'sighandler';
    $SIG{'TERM'} = 'sighandler';

    # prevent termination procedure to run twice     
    if ($cpu == 0) {
	print STDERR "Finishing dirty. Temporary data may not be reusable and should be recalculated.\n";
	print STDERR "You might need to remove it manualy or use the force-option for the next run!\n";
	return;
    }

    # tell the user, we are going to shut down
    print STDERR "\nSignal: $signal caught. Sending the signal again cancels the operation. In this case, blast output will not be reusable!\n\nPrepering for clean termination.\nFetching active threads...\n";

    # don't start new threads while waiting
    $cpu = 0;

    # remove the file lock if you have one
    if ($gotlock == 1) {
    	&fileUnlock;
    }

    # fetch all threads
    # blastoutputs and fasta-databases are written within the threads
    # for everything else the return values do not matter at all if the process is canceled
    if (defined(threads->list())) {
	    foreach my $thread (threads->list()) {
			my @list = threads->list();
	   	      print STDERR " ".(scalar(@list))." threads running, please wait\n";
			$thread->join();
	    }
    }

    # tell the user, we had shut down successfuly
    print STDERR "\nRun aborted.\n";
    exit;
}

